#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy


def bresenham_line(x0: float, y0: float, x1: float, y1: float):
    """
    Bresenham's Line Algorithm
    Produces a list of tuples from start and end

    :param x0: x-coordinate of the start point
    :param y0: y-coordinate of the start point
    :param x1: x-coordinate of the end point
    :param y1: y-coordinate of the end point
    :return: Numpy array of points (x, y) along the line
    """
    points = []
    dx = abs(x1 - x0)
    dy = abs(y1 - y0)
    x, y = x0, y0
    sx = -1 if x0 > x1 else 1
    sy = -1 if y0 > y1 else 1

    if dx > dy:
        err = dx / 2.0
        while x != x1:
            points.append((x, y))
            err -= dy
            if err < 0:
                y += sy
                err += dx
            x += sx
    else:
        err = dy / 2.0
        while y != y1:
            points.append((x, y))
            err -= dx
            if err < 0:
                x += sx
                err += dy
            y += sy

    points.append((x, y))  # Make sure the end point is included
    return numpy.array(points).T



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import LightWave2D


__all__ = [
    'root_path',
    'project_path',
    'doc_path',
    'doc_css_path',
    'version_path',
]

root_path = Path(LightWave2D.__path__[0])

project_path = root_path.parents[0]

doc_path = project_path.joinpath('docs')

doc_css_path = doc_path.joinpath('source/_static/default.css')

version_path = root_path.joinpath('VERSION')


if __name__ == '__main__':
    for path_name in __all__:
        path = locals()[path_name]
        print(path)
        assert path.exists(), f"Path {path_name} do not exists"

# -



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import NoReturn
import numpy
from LightWave2D.grid import Grid
from pydantic.dataclasses import dataclass
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt

config_dict = dict(
    kw_only=True,
    slots=True,
    extra='forbid',
    arbitrary_types_allowed=True
)


@dataclass(config=config_dict)
class PML():
    grid: Grid
    """ The grid of the simulation mesh """
    width: int = 10
    """ Width of the PML region """
    sigma_max: float = 0.045
    """ Adjust sigma_max for better absorption, based on wavelength and PML width """
    order: int = 3
    """ Polynomial order of sigma profile """

    def __post_init__(self):
        self.sigma_x = numpy.zeros((self.grid.n_x, self.grid.n_y))
        self.sigma_y = numpy.zeros((self.grid.n_x, self.grid.n_y))

        for i in range(self.grid.n_x):
            for j in range(self.grid.n_y):
                # Left and right PML regions
                if i < self.width:
                    self.sigma_x[i, j] = self.sigma_max * ((self.width - i) / self.width) ** self.order
                elif i >= self.grid.n_x - self.width:
                    self.sigma_x[i, j] = self.sigma_max * ((i - (self.grid.n_x - self.width - 1)) / self.width) ** self.order

                # Top and bottom PML regions
                if j < self.width:
                    self.sigma_y[i, j] = self.sigma_max * ((self.width - j) / self.width) ** self.order
                elif j >= self.grid.n_y - self.width:
                    self.sigma_y[i, j] = self.sigma_max * ((j - (self.grid.n_y - self.width - 1)) / self.width) ** self.order

    def add_to_ax(self, ax: plt.axis) -> NoReturn:
        cmap = numpy.zeros([256, 4])
        cmap[:, 3] = numpy.linspace(0, 1, 256)
        cmap = ListedColormap(cmap)

        ax.pcolormesh(
            self.grid.x_stamp,
            self.grid.y_stamp,
            self.sigma_y.T + self.sigma_x.T,
            cmap=cmap,
            label='PML'
        )

    def plot(self, unit_size: int = 6) -> NoReturn:
        figsize = int(unit_size), int(unit_size * self.grid.size_y / self.grid.size_x)
        figure, ax = plt.subplots(1, 1, figsize=figsize)

        ax.set_title('FDTD Simulation at time step')
        ax.set_xlabel(r'x position [$\mu$m]')
        ax.set_xlabel(r'y position [$\mu$m]')
        ax.set_aspect('equal')

        self.add_to_ax(ax)

        plt.show()

# -



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Tuple, NoReturn
import numpy as np
from LightWave2D.utils import bresenham_line
from pydantic.dataclasses import dataclass
from LightWave2D.components.base_class import BaseComponent
from matplotlib.path import Path
import shapely.geometry as geo
import matplotlib.pyplot as plt
from LightWave2D.physics import Physics

config_dict = {
    'kw_only': True,
    'slots': True,
    'extra': 'forbid',
    'arbitrary_types_allowed': True
}


@dataclass(kw_only=True, config=config_dict)
class PointSource(BaseComponent):
    """
    Represents a point source in a 2D light wave simulation.

    Attributes:
        wavelength (float): Wavelength of the source.
        position (tuple): Position (x, y) of the source.
        amplitude (float): Amplitude of the electric field, default is 1.0.
        facecolor (str): Color for the source's face, default is 'red'.
        edgecolor (str): Color for the source's edge, default is 'red'.
    """
    wavelength: float
    position: Tuple[float | str, float | str]
    amplitude: float = 1.0
    facecolor: str = 'red'
    edgecolor: str = 'red'

    def __post_init__(self):
        self.frequency = Physics.c / self.wavelength
        self.omega = 2 * np.pi * self.frequency
        x, y = self.position
        self.p0 = self.grid.get_coordinate(x=x, y=y)
        self.polygon = geo.Point(self.p0.x, self.p0.y)
        self.path = Path(self.polygon.coords)

    def add_to_ax(self, ax: plt.Axes) -> NoReturn:
        """
        Add the point source to the provided axis as a scatter point.

        Args:
            ax (plt.Axes): The axis to which the point source will be added.
        """
        ax.scatter(self.p0.x, self.p0.y, color=self.facecolor)

    def add_source_to_field(self, field: np.ndarray, time: float) -> NoReturn:
        """
        Add the source's effect to the simulation field.

        Args:
            field (np.ndarray): The simulation field to which the source's effect will be added.
            time (float): The current simulation time.
        """
        field[self.p0.x_index, self.p0.y_index] += self.amplitude * np.sin(self.omega * time)


@dataclass(kw_only=True, config=config_dict)
class LineSource(BaseComponent):
    """
    Represents a line source in a 2D light wave simulation.

    Attributes:
        wavelength (float): Wavelength of the source.
        point_0 (Tuple[float, float]): Starting position (x, y) of the source.
        point_1 (Tuple[float, float]): Ending position (x, y) of the source.
        amplitude (float): Amplitude of the electric field, default is 1.0.
    """
    wavelength: float
    point_0: Tuple[float | str, float | str]
    point_1: Tuple[float | str, float | str]
    amplitude: float = 1.0
    facecolor: str = 'red'
    edgecolor: str = 'red'

    def __post_init__(self):
        self.frequency = Physics.c / self.wavelength
        self.omega = 2 * np.pi * self.frequency
        self.build_object()

    def build_object(self) -> NoReturn:
        """
        Build the line source object and calculate the line's coordinates.
        """
        self.p0 = self.grid.get_coordinate(x=self.point_0[0], y=self.point_0[1])
        self.p1 = self.grid.get_coordinate(x=self.point_1[0], y=self.point_1[1])

        position = bresenham_line(
            x0=self.p0.x_index,
            y0=self.p0.y_index,
            x1=self.p1.x_index,
            y1=self.p1.y_index
        )

        rows, cols = zip(*position.T)
        self.slice_indexes = rows, cols

        p0 = geo.Point(self.p0.x, self.p0.y)
        p1 = geo.Point(self.p1.x, self.p1.y)
        self.polygon = geo.LineString((p0, p1))

    def add_to_ax(self, ax: plt.axis) -> NoReturn:
        """
        Add the scatterer to the provided axis as a circle.

        Args:
            ax (Axis): The axis to which the scatterer will be added.
        """
        ax.plot(self.polygon.xy[0], self.polygon.xy[1], color=self.facecolor)

    def add_source_to_field(self, field: np.ndarray, time: float) -> NoReturn:
        """
        Add the source's effect to the simulation field.

        Args:
            field (np.ndarray): The simulation field to which the source's effect will be added.
            time (float): The current simulation time.
        """
        rows, cols = self.slice_indexes
        field[rows, cols] += self.amplitude * np.sin(self.omega * time)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Tuple, Union, NoReturn
from dataclasses import field
import numpy
from LightWave2D.grid import Grid
from MPSPlots.render2D import SceneList
import shapely.geometry as geo
from matplotlib.path import Path
from pydantic.dataclasses import dataclass
from LightWave2D.components.base_class import BaseComponent
import matplotlib.pyplot as plt

config_dict = dict(
    kw_only=True,
    slots=True,
    extra='forbid',
    arbitrary_types_allowed=True
)


@dataclass(kw_only=True, config=config_dict)
class PointDetector(BaseComponent):
    """
    Represents a point detector within a simulation grid.

    Attributes:
        grid (Grid): The simulation mesh grid.
        position (Tuple[float | str, float | str]): The (x, y) coordinates of the point detector on the grid.
        data (numpy.ndarray): The data collected by the detector over time.
    """
    position: Union[Tuple[float | str, float | str]]
    data: numpy.ndarray = field(init=False)

    def __post_init__(self):
        """
        Initialize the PointDetector by converting the given position into
        grid coordinates and initializing the data array.
        """
        x, y = self.position

        self.p0 = self.grid.get_coordinate(x=x, y=y)

        self.polygon = geo.Point(self.p0.x, self.p0.y)

        self.path = Path(self.polygon.coords)

        self.data = numpy.zeros(self.grid.n_steps)

    def update_data(self, field: numpy.ndarray) -> NoReturn:
        """
        Update the detector data based on the provided field values.

        Parameters:
            field (numpy.ndarray): The field values to update the detector data.
        """
        self.data = field[:, self.p0.x_index, self.p0.y_index]

    def plot_data(self) -> SceneList:
        """
        Plot the detector data over time.

        Returns:
            SceneList: A scene list containing the plot.
        """
        scene = SceneList()
        ax = scene.append_ax()
        ax.add_line(x=self.grid.time_stamp, y=self.data)
        return scene

    def add_to_ax(self, ax: plt.axis) -> NoReturn:
        """
        Add the scatterer to the provided axis as a circle.

        Args:
            ax (Axis): The axis to which the scatterer will be added.
        """
        ax.scatter(
            self.p0.x,
            self.p0.y,
            color=self.facecolor
        )

# -



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Tuple, Union, NoReturn
from pydantic.dataclasses import dataclass
from LightWave2D.grid import Grid
from LightWave2D.components.base_class import BaseComponent
from matplotlib.path import Path
import shapely.geometry as geo

# Configuration dictionary for dataclasses
config_dict = {
    'kw_only': True,
    'extra': 'forbid',
    'slots': True,
    'arbitrary_types_allowed': True
}


@dataclass(config=config_dict)
class Grating(BaseComponent):
    """
    Represents a grating in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): Starting position of the grating.
        epsilon_r (float): Relative permittivity inside the grating.
        period (float): Period of the grating.
        duty_cycle (float): Duty cycle of the grating (fraction of period occupied by the grating material).
        num_periods (int): Number of grating periods.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    period: float
    duty_cycle: float
    num_periods: int

    def compute_polygon(self) -> NoReturn:
        """
        Compute the path of the grating.
        """
        bars = []
        for i in range(self.num_periods):
            x_start = self.coordinate.x + i * self.period
            x_end = x_start + self.duty_cycle * self.period
            bar = geo.box(x_start, self.coordinate.y - 0.5, x_end, self.coordinate.y + 0.5)
            bars.append(bar)

        self.polygon = geo.MultiPolygon(bars)


@dataclass(config=config_dict)
class RingResonator(BaseComponent):
    """
    Represents a ring resonator in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): Starting position of the resonator (center of the ring).
        epsilon_r (float): Relative permittivity inside the resonator.
        inner_radius (float): Inner radius of the ring.
        width (float): Outer radius of the ring.
    """
    grid: Grid
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    inner_radius: float
    width: float

    def compute_polygon(self) -> Path:
        """
        Compute the path of the ring resonator.
        """
        inner_circle = geo.Point(self.coordinate.x, self.coordinate.y).buffer(self.inner_radius)
        outer_circle = geo.Point(self.coordinate.x, self.coordinate.y).buffer(self.inner_radius + self.width)
        self.polygon = outer_circle.difference(inner_circle)


@dataclass(config=config_dict)
class Lense(BaseComponent):
    """
    Represents a lens in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): Starting position of the lens (center of the lens).
        epsilon_r (float): Relative permittivity inside the lens.
        radius (float): Radius of the lens.
        curvature (float): Curvature of the lens (positive for convex, negative for concave).
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    width: float
    curvature: float

    def compute_polygon(self) -> Path:
        """
        Compute the path of the lens.
        """
        # Create two arcs for the lens shape
        x0 = self.coordinate.x + self.curvature - self.width / 2
        x1 = self.coordinate.x - self.curvature + self.width / 2

        arc3 = geo.Point(x0, self.coordinate.y).buffer(self.curvature, resolution=100)
        arc4 = geo.Point(x1, self.coordinate.y).buffer(self.curvature, resolution=100)

        self.polygon = arc3.intersection(arc4)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy
from dataclasses import dataclass


@dataclass(frozen=True)
class Physics():
    c: float = 3e8
    """ Speed of light in vacuum, m/s """
    epsilon_0: float = 8.854e-12
    """ Permittivity of free space """
    mu_0: float = 4 * numpy.pi * 1e-7
    """ Permeability of free space """

# -



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Tuple, Union, NoReturn
from pydantic.dataclasses import dataclass
import numpy as np
from LightWave2D.grid import Grid
from LightWave2D.components.base_class import BaseComponent
from matplotlib.path import Path
import shapely.geometry as geo
from matplotlib import patches
from shapely.affinity import scale

# Configuration dictionary for dataclasses
config_dict = {
    'kw_only': True,
    'extra': 'forbid',
    'slots': True,
    'arbitrary_types_allowed': True
}


@dataclass(config=config_dict)
class Square(BaseComponent):
    """
    Represents a square scatterer in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): Starting position of the scatterer (center of the square).
        epsilon_r (float): Relative permittivity inside the scatterer.
        side_length (float): Side length of the square scatterer.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    side_length: float

    def compute_polygon(self) -> Path:
        """
        Compute the path of the square scatterer.
        """
        half_side = self.side_length / 2
        x_start, x_end = self.coordinate.x - half_side, self.coordinate.x + half_side
        y_start, y_end = self.coordinate.y - half_side, self.coordinate.y + half_side

        self.polygon = geo.Polygon([
            (x_start, y_start),
            (x_start, y_end),
            (x_end, y_end),
            (x_end, y_start)
        ])


@dataclass(config=config_dict)
class Circle(BaseComponent):
    """
    Represents a circular scatterer in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): Starting position of the scatterer.
        epsilon_r (float): Relative permittivity inside the scatterer.
        radius (float): Radius of the scatterer.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    radius: float

    def compute_polygon(self) -> Path:
        """
        Compute the path of the circular scatterer.
        """
        self.polygon = geo.Point(self.coordinate.x, self.coordinate.y).buffer(self.radius)


@dataclass(config=config_dict)
class Triangle(BaseComponent):
    """
    Represents a triangular scatterer in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): Starting position of the scatterer (center of the triangle).
        epsilon_r (float): Relative permittivity inside the scatterer.
        side_length (float): Side length of the triangle.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    epsilon_r: float
    side_length: float

    def compute_polygon(self) -> NoReturn:
        """
        Compute the path of the triangular scatterer.
        """
        height = (np.sqrt(3) / 2) * self.side_length

        # Vertices of an equilateral triangle centered at (coordinate.x, coordinate.y)
        p0 = (self.coordinate.x - self.side_length / 2, self.coordinate.y - height / 3)
        p1 = (self.coordinate.x + self.side_length / 2, self.coordinate.y - height / 3)
        p2 = (self.coordinate.x, self.coordinate.y + 2 * height / 3)

        self.polygon = geo.Polygon([p0, p1, p2])


@dataclass(config=config_dict)
class Ellipse(BaseComponent):
    """
    Represents an elliptical scatterer in a simulation grid.

    Attributes:
        position (Tuple[Union[float, str], Union[float, str]]): Starting position of the scatterer.
        width (float): Width of the ellipse.
        height (float): Height of the ellipse.
        epsilon_r (float): Relative permittivity inside the scatterer.
        resolution (float): Resolution of the ellipse.
    """
    position: Tuple[Union[float, str], Union[float, str]]
    width: float
    height: float
    epsilon_r: float
    resolution: float = 100

    def compute_polygon(self) -> Path:
        """
        Compute the path of the elliptical scatterer.
        """
        ellipse_patch = patches.Ellipse(
            (self.coordinate.x, self.coordinate.y),
            width=self.resolution,
            height=self.resolution * self.height / self.width,
        )

        vertices = ellipse_patch.get_verts()
        polygon = geo.Polygon(vertices)

        self.polygon = scale(polygon, xfact=self.width / self.resolution, yfact=self.width / self.resolution)



"""
Experiment: scattering
======================

"""

# %%
# Importing the package
from LightWave2D.grid import Grid
from LightWave2D.experiment import Experiment
from MPSPlots.colormaps import polytechnique

# %%
# We define here the grid on which to build the experiemnt
grid = Grid(
    resolution=0.1e-6,
    size_x=50e-6,
    size_y=30e-6,
    n_steps=800
)

experiment = Experiment(grid=grid)


scatterer = experiment.add_ring_resonator(
    position=('35%', '50%'),
    epsilon_r=1.5,
    # period=3e-6,
    inner_radius=4e-6,
    width=2e-6
)

scatterer = experiment.add_ring_resonator(
    position=('60%', '50%'),
    epsilon_r=1.5,
    # period=3e-6,
    inner_radius=4e-6,
    width=2e-6
)


# scatterer = experiment.add_grating(
#     position=('25%', '50%'),
#     epsilon_r=2,
#     period=3e-6,
#     duty_cycle=1e-6,
#     num_periods=5
# )

scatterer = experiment.add_lense(
    position=('25%', '50%'),
    epsilon_r=2,
    width=5e-6,
    curvature=20e-6
)

# # %%
# # We add a circular scatterer
# scatterer = experiment.add_square(
#     position=('25%', '20%'),
#     epsilon_r=2,
#     side_length=3e-6
# )

# # scatterer.plot()

# # %%
# # We add a circular scatterer
# scatterer = experiment.add_ellipse(
#     position=('25%', '70%'),
#     epsilon_r=2,
#     width=5e-6,
#     height=10e-6,
#     angle=10
# )


# scatterer = experiment.add_triangle(
#     position=('55%', '50%'),
#     epsilon_r=2,
#     side_length=4e-6,
#     # angle=10
# )

# scatterer.plot()

# %%
# We add a line source
source = experiment.add_point_source(
    wavelength=1550e-9,
    position=('20%', '50%'),
    amplitude=100,
)

# %%
# We add a line source
# source = experiment.add_line_source(
#     wavelength=1550e-9,
#     point_0=('10%', '40%'),
#     point_1=('10%', '60%'),
#     amplitude=10,
# )


# %%
# We add a perfectly matched layer to avoid reflection at the boundary of the mesh
pml = experiment.add_pml(order=1, width=70, sigma_max=5000)
# pml.plot()


# %%
# We add a detector
# detector = experiment.add_point_detector(position=(25e-6, 'center'))

# %%
# Plotting of the whole experiemnt setup
experiment.plot()

experiment.run_fdtd()

experiment.plot_frame(frame_number=-1, scale_max=8)

animation = experiment.render_propagation(skip_frame=5, colormap=polytechnique.red_black_blue)

animation.save('./tests.gif', writer='Pillow', fps=10)

# -



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import NoReturn
from pydantic.dataclasses import dataclass
import numpy
from LightWave2D.physics import Physics
import matplotlib.pyplot as plt
from matplotlib.patches import PathPatch
import matplotlib as mpl
import numpy as np
from matplotlib.path import Path
from matplotlib.collections import PatchCollection
from LightWave2D.grid import Grid


# Configuration dictionary for dataclasses
config_dict = {
    'kw_only': True,
    'extra': 'forbid',
    'slots': True,
    'arbitrary_types_allowed': True
}


@dataclass(kw_only=True, config=config_dict)
class BaseComponent():
    """
    Represents an elliptical scatterer in a simulation grid.

    Attributes:
        grid (Grid): The grid of the simulation mesh.
        angle (float): Rotation angle of the ellipse.
        facecolor (str): Color of the scatterer face.
        edgecolor (str): Color of the scatterer edge.
    """
    grid: Grid
    facecolor: str = 'lightblue'
    edgecolor: str = 'blue'
    alpha: float = 0.3
    rotation: float = 0

    def __post_init__(self):
        x0, y0 = self.position
        self.coordinate = self.grid.get_coordinate(x=x0, y=y0)

        self.build_object()

    def build_object(self) -> NoReturn:
        """
        Build the permittivity mesh for the square scatterer.
        """
        self.epsilon_r_mesh = numpy.ones(self.grid.shape)  # Background permittivity

        self.compute_polygon()

        self.path = Path(self.polygon.exterior.coords)

        self.path = self.path.transformed(mpl.transforms.Affine2D().rotate_around(self.coordinate.x, self.coordinate.y, self.rotation))

        x_mesh, y_mesh = numpy.meshgrid(self.grid.x_stamp, self.grid.y_stamp)

        coordinates = numpy.c_[x_mesh.T.flatten(), y_mesh.T.flatten()]

        self.idx = self.path.contains_points(coordinates).astype(bool).reshape(self.epsilon_r_mesh.shape)

        self.epsilon_r_mesh[self.idx] = self.epsilon_r

    def add_to_ax(self, ax: plt.axis) -> PatchCollection:
        """
        Add the scatterer to the provided axis as a circle.

        Args:
            ax (Axis): The axis to which the scatterer will be added.
        """
        path = Path.make_compound_path(
            Path(np.asarray(self.polygon.exterior.coords)[:, :2]),
            *[Path(np.asarray(ring.coords)[:, :2]) for ring in self.polygon.interiors])

        patch = PathPatch(path, color=self.facecolor, edgecolor=self.edgecolor, alpha=0.4)
        collection = PatchCollection([patch], color=self.facecolor, edgecolor=self.edgecolor, alpha=self.alpha)

        ax.add_collection(collection, autolim=True)
        ax.autoscale_view()
        return collection

    def plot(self) -> NoReturn:
        figure, ax = plt.subplots(1, 1, figsize=(6, 6))
        ax.set_title('FDTD Simulation')
        ax.set_xlabel(r'x position [$\mu$m]')
        ax.set_xlabel(r'y position [$\mu$m]')
        ax.set_aspect('equal')

        self.add_to_ax(ax)
        ax.autoscale_view()

        plt.show()

    def interpret_position_to_index(self) -> tuple:
        """
        Interprets and returns the value of positions x and y of the waveguide

        :returns:   The index of width and height
        :rtype:     int
        """
        x, y = self.position

        if isinstance(x, str):
            x_index = self.grid.string_to_index(x, axis='x')
        else:
            x_index = self.grid.position_to_index(x, axis='x')

        if isinstance(y, str):
            y_index = self.grid.string_to_index(y, axis='y')
        else:
            y_index = self.grid.position_to_index(y, axis='y')

        return x_index, y_index

    def interpret_position_to_position(self) -> tuple:
        """
        Interprets and returns the value of positions x and y of the waveguide

        :returns:   The index of width and height
        :rtype:     int
        """
        x, y = self.position

        coordinate = self.grid.get_coordinate(x=x, y=y)

        return coordinate.x, coordinate.y

    def add_to_mesh(self, epsilon_r_mesh: numpy.ndarray) -> None:
        epsilon_r_mesh += self.epsilon_r_mesh

    def add_non_linear_effect_to_field(self, field: numpy.ndarray) -> None:
        chi_2 = 1e10

        field += self.idx * self.grid.dt**2 / (self.epsilon_r_mesh * Physics.epsilon_0 * Physics.mu_0) * chi_2 * field ** 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Union, Optional
import numpy
from LightWave2D.physics import Physics
from pydantic.dataclasses import dataclass

config_dict = dict(
    kw_only=True,
    slots=True,
    extra='forbid'
)


class NameSpace:
    """
    A class to dynamically create attributes from keyword arguments.
    """

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


@dataclass(config=config_dict)
class Grid:
    """
    Represents a 2D simulation grid with specified dimensions and time steps.

    Attributes:
        resolution (float): Spatial resolution of the grid in meters per cell.
        size_x (float): Size of the grid in the x-direction in meters.
        size_y (float): Size of the grid in the y-direction in meters.
        n_steps (int): Number of time steps for the simulation (default is 200).
    """
    resolution: float
    size_x: float
    size_y: float
    n_steps: int = 200

    def __post_init__(self):
        self.n_x = int(self.size_x / self.resolution)
        self.n_y = int(self.size_y / self.resolution)
        self.dx = self.size_x / self.n_x  # Should be approximately equal to the resolution
        self.dy = self.size_y / self.n_y  # Should be approximately equal to the resolution
        self.dt = 1 / (Physics.c * numpy.sqrt(1 / self.dx**2 + 1 / self.dy**2))  # Time step size using Courant condition

        self.shape = (self.n_x, self.n_y)
        self.time_stamp = numpy.arange(self.n_steps) * self.dt
        self.x_stamp = numpy.arange(self.n_x) * self.dx
        self.y_stamp = numpy.arange(self.n_y) * self.dy

    def get_distance_grid(self, x0: float = 0, y0: float = 0) -> numpy.ndarray:
        """
        Compute the distance grid from a given point (x0, y0).

        Args:
            x0 (float): x-coordinate of the reference point (default is 0).
            y0 (float): y-coordinate of the reference point (default is 0).

        Returns:
            numpy.ndarray: A 2D array of distances from the reference point.
        """
        x_mesh, y_mesh = numpy.meshgrid(self.x_stamp, self.y_stamp)
        distance_mesh = numpy.sqrt((x_mesh - x0)**2 + (y_mesh - y0)**2)
        return distance_mesh

    def get_coordinate(self, x: Optional[Union[float, str]] = None, y: Optional[Union[float, str]] = None) -> NameSpace:
        """
        Get the coordinate and index for a given position in the grid.

        Args:
            x (float | str): x-coordinate or position string ('left', 'center', 'right').
            y (float | str): y-coordinate or position string ('bottom', 'center', 'top').

        Returns:
            NameSpace: An object containing the coordinates and indices.
        """
        coordinate = NameSpace()

        if isinstance(x, str):
            x = self.parse_x_position(x)
        if isinstance(y, str):
            y = self.parse_y_position(y)

        if x is not None:
            x = numpy.clip(x, self.x_stamp[0], self.x_stamp[-1])
            coordinate.x = x
            coordinate.x_index = int(x / self.dx)

        if y is not None:
            y = numpy.clip(y, self.y_stamp[0], self.y_stamp[-1])
            coordinate.y = y
            coordinate.y_index = int(y / self.dy)

        return coordinate

    def parse_y_position(self, value: Union[str, float]) -> float:
        """
        Convert a position string to a y-coordinate.

        Args:
            position_string (str): Position string ('bottom', 'center', 'top').

        Returns:
            float: Corresponding y-coordinate.
        """
        if isinstance(value, str):
            value = value.lower()
            if '%' in value:
                percentage = float(value.strip('%')) / 100.0
                delta = self.y_stamp[-1] - self.y_stamp[0]

                return self.y_stamp[0] + percentage * delta

            assert value in ['bottom', 'center', 'top'], f"Invalid position: {value}. Valid inputs are ['bottom', 'center', 'top']."
            match value:
                case 'bottom':
                    return self.y_stamp[0]
                case 'center':
                    return numpy.mean(self.y_stamp)
                case 'top':
                    return self.y_stamp[-1]

        return value

    def parse_x_position(self, value: Union[str, float]) -> float:
        """
        Convert a position string to an x-coordinate.

        Args:
            position_string (str): Position string ('left', 'center', 'right').

        Returns:
            float: Corresponding x-coordinate.
        """
        if isinstance(value, str):
            value = value.lower()
            if '%' in value:
                percentage = float(value.strip('%')) / 100.0
                delta = self.x_stamp[-1] - self.x_stamp[0]

                return self.x_stamp[0] + percentage * delta

            assert value in ['right', 'center', 'left'], f"Invalid position: {value}. Valid inputs are ['right', 'center', 'left']."
            match value:
                case 'right':
                    return self.x_stamp[-1]
                case 'center':
                    return numpy.mean(self.x_stamp)
                case 'left':
                    return self.x_stamp[0]

        return value




#!/usr/bin/env python
# -*- coding: utf-8 -*-


from typing import Tuple, NoReturn, Optional, Union
import numpy
from LightWave2D.physics import Physics
from LightWave2D.grid import Grid
from LightWave2D.components.basic import Circle, Square, Ellipse, Triangle
from LightWave2D.components.complex import Lense, Grating, RingResonator
from LightWave2D.source import PointSource, LineSource
from LightWave2D.detector import PointDetector
from LightWave2D.pml import PML
from MPSPlots import colormaps
from MPSPlots.render2D import SceneList
import matplotlib.animation as animation
from pydantic.dataclasses import dataclass
import matplotlib.pyplot as plt

config_dict = dict(
    kw_only=True,
    slots=True,
    extra='forbid',
    arbitrary_types_allowed=True
)


@dataclass(config=config_dict)
class Experiment:
    """Class representing an FDTD simulation experiment."""

    grid: Grid
    """The grid of the simulation mesh."""

    def __post_init__(self):
        self.sources = []
        self.components = []
        self.detectors = []
        self.Ez_t = numpy.zeros((self.grid.n_steps, *self.grid.shape))
        self.epsilon = numpy.ones(self.grid.shape) * Physics.epsilon_0
        self.pml = None

    def get_gradient(self, field: numpy.ndarray, axis: str) -> numpy.ndarray:
        """
        Compute the gradient of the field along the specified axis.

        Args:
            field (numpy.ndarray): The field to compute the gradient for.
            axis (str): The axis along which to compute the gradient ('x' or 'y').

        Returns:
            numpy.ndarray: The computed gradient.
        """
        if axis == 'x':
            gradient = numpy.diff(field, axis=0) / self.grid.dx
        elif axis == 'y':
            gradient = numpy.diff(field, axis=1) / self.grid.dy
        else:
            raise ValueError("Axis must be 'x' or 'y'.")
        return gradient

    def plot(self, unit_size: int = 6, colormap: Optional[Union[str, object]] = 'Blues') -> NoReturn:
        """
        Generates a plot of the FDTD simulation setup using a specified colormap.

        This function initializes a plotting scene and adds elements of the FDTD simulation
        such as components, sources, detectors, and optionally the PML layers. It adjusts
        the axis properties and attaches a colorbar based on the selected colormap.

        Args:
            colormap (Optional[Union[str, object]]): The colormap to use for the plot. If 'None' is passed, a default dark-themed colormap is used. Default is 'Blues'.

        Returns:
            SceneList: A SceneList object that contains the constructed plot.
        """
        figure, ax = self.get_figure_ax(unit_size=unit_size)

        # Add PML layers to the plot if present
        if self.pml:
            self.pml.add_to_ax(ax)

        for component in [*self.components, *self.sources, *self.detectors]:
            component.add_to_ax(ax)

        ax.legend()
        ax.autoscale_view()

        plt.show()

    def add_to_component(function):
        def wrapper(self, **kwargs):
            component = function(self, **kwargs)
            self.components.append(component)
            return component
        return wrapper

    def add_to_source(function):
        def wrapper(self, **kwargs):
            source = function(self, **kwargs)
            self.sources.append(source)
            return source
        return wrapper

    def add_to_detector(function):
        def wrapper(self, **kwargs):
            detector = function(self, **kwargs)
            self.detectors.append(detector)
            return detector
        return wrapper

    def add_pml(self, **kwargs) -> PML:
        """Add a Perfectly Matched Layer (PML) to the simulation."""
        self.pml = PML(grid=self.grid, **kwargs)
        return self.pml

    @add_to_component
    def add_circle(self, **kwargs) -> Circle:
        """General method to add a component to the simulation."""
        return Circle(grid=self.grid, **kwargs)

    @add_to_component
    def add_ellipse(self, **kwargs) -> Ellipse:
        """General method to add a component to the simulation."""
        return Ellipse(grid=self.grid, **kwargs)

    @add_to_component
    def add_square(self, **kwargs) -> Square:
        """General method to add a component to the simulation."""
        return Square(grid=self.grid, **kwargs)

    @add_to_component
    def add_triangle(self, **kwargs) -> Square:
        """General method to add a component to the simulation."""
        return Triangle(grid=self.grid, **kwargs)

    @add_to_component
    def add_lense(self, **kwargs) -> Square:
        """General method to add a component to the simulation."""
        return Lense(grid=self.grid, **kwargs)

    @add_to_component
    def add_grating(self, **kwargs) -> Square:
        """General method to add a component to the simulation."""
        return Grating(grid=self.grid, **kwargs)

    @add_to_component
    def add_ring_resonator(self, **kwargs) -> Square:
        """General method to add a component to the simulation."""
        return RingResonator(grid=self.grid, **kwargs)

    @add_to_source
    def add_point_source(self, **kwargs) -> PointSource:
        """General method to add a source to the simulation."""
        return PointSource(grid=self.grid, **kwargs)

    @add_to_source
    def add_line_source(self, **kwargs) -> LineSource:
        """General method to add a source to the simulation."""
        return LineSource(grid=self.grid, **kwargs)

    @add_to_detector
    def add_point_detector(self, **kwargs) -> PointDetector:
        """General method to add a detector to the simulation."""
        return PointDetector(grid=self.grid, **kwargs)

    def get_sigma(self) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        Retrieve the sigma values for the PML.

        Returns:
            tuple: Sigma values for x and y directions.
        """
        if self.pml is not None:
            sigma_x, sigma_y = self.pml.sigma_x, self.pml.sigma_y
        else:
            sigma_x = sigma_y = numpy.zeros(self.grid.shape)
        return sigma_x, sigma_y

    def get_epsilon(self) -> numpy.ndarray:
        """
        Construct the epsilon mesh with contributions from all components.

        Returns:
            numpy.ndarray: The epsilon mesh.
        """
        epsilon_r_mesh = numpy.ones(self.grid.shape) * Physics.epsilon_0
        for component in self.components:
            component.add_to_mesh(epsilon_r_mesh)
        return epsilon_r_mesh * Physics.epsilon_0

    def get_field_yee_gradient(self, field: numpy.ndarray) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        Calculate the Yee grid gradient of the field.

        Args:
            field (numpy.ndarray): The field to calculate the gradient for.

        Returns:
            tuple: Gradients along x and y directions.
        """
        d_dx = (field[1:, :] - field[:-1, :]) / self.grid.dx
        d_dy = (field[:, 1:] - field[:, :-1]) / self.grid.dy
        return d_dx, d_dy

    def run_fdtd(self) -> NoReturn:
        """Run the FDTD simulation."""
        Ez = numpy.zeros(self.grid.shape)
        Hx = numpy.zeros(self.grid.shape)
        Hy = numpy.zeros(self.grid.shape)

        sigma_x, sigma_y = self.get_sigma()
        epsilon = self.get_epsilon()
        mu_factor = self.grid.dt / Physics.mu_0
        eps_factor = self.grid.dt / epsilon

        for iteration, t in enumerate(self.grid.time_stamp):

            dEz_dx, dEz_dy = self.get_field_yee_gradient(Ez)
            Hx[:, :-1] -= mu_factor * dEz_dy * (1 - sigma_y[:, :-1] * mu_factor / 2)
            Hy[:-1, :] += mu_factor * dEz_dx * (1 - sigma_x[:-1, :] * mu_factor / 2)

            dHy_dx = (Hy[1:-1, 1:-1] - Hy[:-2, 1:-1]) / self.grid.dx
            dHx_dy = (Hx[1:-1, 1:-1] - Hx[1:-1, :-2]) / self.grid.dy

            Ez[1:-1, 1:-1] += eps_factor[1:-1, 1:-1] * (dHy_dx - dHx_dy)

            for component in self.components:
                component.add_non_linear_effect_to_field(Ez)

            Ez *= (1 - (sigma_x + sigma_y) * eps_factor / 2)

            for source in self.sources:
                source.add_source_to_field(Ez, time=t)

            self.Ez_t[iteration] = Ez

        for detector in self.detectors:
            detector.update_data(field=self.Ez_t)

    def plot_frame(self, frame_number: int, scale_max: float = 5, unit_size: int = 6, colormap: Optional[Union[str, object]] = colormaps.polytechnique.blue_black_red) -> SceneList:
        """
        Creates a plot of a specific frame from the FDTD simulation.

        This method generates a visualization of the electric field distribution along with all
        relevant simulation components, sources, and detectors for a given frame.

        Args:
            frame_number (int): The index of the frame to be visualized.
            colormap (Optional[Union[str, object]]): The colormap used for the visualization.
                Defaults to a blue-black-red colormap from Polytechnique collection.

        Returns:
            SceneList: A SceneList object that contains the constructed plot.
        """
        figure, ax = self.get_figure_ax(unit_size=unit_size)

        image = ax.pcolormesh(
            self.grid.x_stamp,
            self.grid.y_stamp,
            self.Ez_t[frame_number].T,
            cmap=colormap
        )

        for component in [*self.components, *self.sources, *self.detectors]:
            component.add_to_ax(ax)

        vmin, vmax = image.get_clim()

        max_diff = max(abs(vmin), abs(vmax)) / scale_max

        image.set_clim([-max_diff, max_diff])

        plt.show()

    def save_frame_as_image(self, frame_number: int, filename: str, scale_max: float = 5, unit_size: int = 6, dpi: int = 200, colormap: Optional[Union[str, object]] = colormaps.polytechnique.blue_black_red) -> None:
        """
        Saves a specific frame from the FDTD simulation as an image file at the specified resolution.

        This method invokes the plot_frame function to generate a visual representation
        of a specific frame and then saves it as an image file using the given colormap.

        Args:
            frame_number (int): The index of the frame to be saved.
            filename (str): The file path where the image will be saved.
            dpi (int): The resolution of the saved image in dots per inch.
            colormap (Optional[Union[str, object]]): The colormap used for visualizing the data.
                Defaults to a predefined blue-black-red colormap.

        Returns:
            None: This method does not return any value, but saves an image file.
        """
        figure, ax = self.get_figure_ax(unit_size=unit_size)

        image = ax.pcolormesh(
            self.grid.x_stamp,
            self.grid.y_stamp,
            self.Ez_t[frame_number].T,
            cmap=colormap
        )

        for component in [*self.components, *self.sources, *self.detectors]:
            component.add_to_ax(ax)

        vmin, vmax = image.get_clim()

        max_diff = max(abs(vmin), abs(vmax)) / scale_max

        image.set_clim([-max_diff, max_diff])

        plt.savefig(filename, dpi=dpi)

    def get_figure_ax(self, unit_size: int = 6) -> Tuple:
        figsize = int(unit_size), int(unit_size * self.grid.size_y / self.grid.size_x)
        figure, ax = plt.subplots(1, 1, figsize=figsize)

        ax.set_title('FDTD Simulation')
        ax.set_xlabel(r'x position [$\mu$m]')
        ax.set_ylabel(r'y position [$\mu$m]')
        ax.set_aspect('equal')

        return figure, ax

    def render_propagation(self, skip_frame: int = 10, scale_max: float = 5, unit_size: int = 6, colormap: Optional[Union[str, object]] = colormaps.blue_black_red) -> animation.FuncAnimation:
        """
        Renders the propagation of a field as an animation using matplotlib.

        This method sets up the scene, defines the initial conditions of the field,
        and creates an animation showing the evolution of the field over time.

        Args:
            skip_frame (int): The number of time steps to skip between frames in the animation.
            colormap (Optional[Union[str, object]]): The colormap used for visualizing the data. Defaults to a predefined blue-black-red colormap.
            unit_size (float): The base unit size for scaling the plot elements.

        Returns:
            animation.FuncAnimation: The animation object that can be displayed or saved.
        """
        figure, ax = self.get_figure_ax(unit_size=unit_size)

        # Initialize the field display
        initial_field = numpy.zeros(self.Ez_t[0].shape).T
        field_artist = ax.pcolormesh(
            self.grid.x_stamp,
            self.grid.y_stamp,
            initial_field,
            cmap=colormap
        )

        # Store all artists for updating
        artist_list = [field_artist]

        # Add other components to the axis and their artists to the list
        for component in self.components:
            artist_list.append(component.add_to_ax(ax))

        # Set color limits based on the maximum field amplitude
        max_amplitude = abs(self.Ez_t).max() / scale_max
        field_artist.set_clim(vmin=-max_amplitude, vmax=max_amplitude)

        def update(frame) -> list:
            """
            Update function for the animation; called for each frame.
            """
            field_t = self.Ez_t[frame].T
            field_artist.set_array(field_t)
            return [artist for artist in artist_list]

        def init_func():
            """
            Initialization function for the animation; called at the start.
            """
            ax.set_xticks([])
            ax.set_yticks([])
            return field_artist,

        # Create and return the animation object
        render = animation.FuncAnimation(
            fig=figure,
            func=update,
            frames=numpy.arange(0, self.grid.n_steps, skip_frame),
            interval=0.2,
            blit=True,
            init_func=init_func
        )

        return render
